/******************************************/
/*                                        */
/* HITTITE: Stellex YIG synth test bed    */
/* controller for Hittite VCO chip        */
/*                                        */
/* Compile with MSVC for Win32            */
/*                                        */
/* jmiles@pop.net                         */
/*                                        */
/******************************************/

#include <math.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <direct.h>
#include <limits.h>
#include <assert.h>
#include <time.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>

#include "pt_ioctl.c"   // Port access library for Win2K/XP

#define VERSION "1.00"

//
// Include platform-independent typedefs and synthesizer driver class
//

#include "typedefs.h"
#include "pllctrl.cpp"

//
// TRUE to make Tek 49x/275x or HP 8566B spectrum analyzer follow the 
// output signal (requires National Instruments GPIB card)
//

#define TEK_49X_CONTROL 0
#define HP_8566_CONTROL 1

//
// TRUE to wait for PLL lock line to become valid after sending frequency command
//

#define WAIT_FOR_LOCK 1

//
// Set to '0' to invert lock-detect signal (compensates for 
// inversion in level-shifter required with 3.3V ADF4107)
//
// Should normally be '1' for other chips
//

#define LD_ACTIVE_LEVEL 1

//
// Synthesizer output range and power-up default
//

#define FOUT_MIN        7850000000
#define FOUT_MAX        8200000000

#define FOUT_DEFAULT    8000000000  //9000000000

//
// Synthesizer's clock
//

#define REF_CLOCK_HZ    80000000

//
// Synthesizer's onboard prescaler factor (prior to the LMX2326 chip)
//

#define PRESCALE_FACTOR 4

//
// The step size is the PLL's comparison frequency.  It must be an integer factor 
// of the clock frequency.  The overall synthesizer circuit will tune in
// steps 4x this size.

#define STEP_SIZE_HZ 2000000

//
// Globals
//

PLL *synth = NULL;      // Synthesizer object under test

U32 out_port;           // LPT port base address
S32 SA_update;          // TRUE to follow output with spectrum analyzer 

/******************************************/
#if TEK_49X_CONTROL
#define ANALYZER_SUPPORT 1

//
// Tek 49x spectrum-analyzer support
//
// Note: Assumes analyzer is at GPIB address 3!
//

#include "gpiblib.h"

void WINAPI GPIB_error(C8 *msg, S32 ibsta, S32 iberr, S32 ibcntl)
{
   printf("\n\n%s",msg);

   exit(1);
}

void GPIB_update(S64 cur_freq)
{
   GPIB_connect(3,
                GPIB_error);

   C8 buffer[80];
   sprintf(buffer,"FREQ %I64d",cur_freq);

   GPIB_write(buffer);

   GPIB_disconnect();
}

void GPIB_plot(S64 cur_freq)
{
   GPIB_connect(3,
                GPIB_error);

   GPIB_write("SIGSWP");

   C8 buffer[80];
   sprintf(buffer,"FREQ %I64d",cur_freq);
   GPIB_write(buffer);

   GPIB_write("SIGSWP");
   GPIB_write("WAIT");

   GPIB_write("PTYPE HP7470");      // select HP7470 plotter compatibility
   GPIB_write("GRAT ON");           // turn on graticule illumination

   sprintf(buffer,"%I64d (%d).plt",cur_freq,synth->current_CM);

   FILE *out = fopen(buffer,"wt");

   fprintf(out, "%s\n", 
      GPIB_query("plot?"));         // send result of PLOT? to stdout

   fclose(out);

   GPIB_write("TRIG FRERUN");
   GPIB_disconnect();
}

void shutdown(void)
{
   GPIB_disconnect();
   ClosePortTalk();
}

#elif HP_8566_CONTROL
#define ANALYZER_SUPPORT 1

//
// HP 856x spectrum-analyzer support
//
// Note: Assumes analyzer is at GPIB address 18!
//

#include "gpiblib.h"

void WINAPI GPIB_error(C8 *msg, S32 ibsta, S32 iberr, S32 ibcntl)
{
   printf("\n\n%s",msg);

   exit(1);
}

void GPIB_update(S64 cur_freq)
{
   GPIB_connect(18,
                GPIB_error,        
                0,
                TIMEOUT_T3s,
                5);

   C8 buffer[80];
   sprintf(buffer,"CF %I64d;\r\n",cur_freq);

   GPIB_write(buffer);

   GPIB_disconnect();
}

void GPIB_plot(S64 cur_freq)
{
}

void shutdown(void)
{
   GPIB_disconnect();
   ClosePortTalk();
}

#else
#define ANALYZER_SUPPORT 0

//
// No spectrum-analyzer support
//

void WINAPI GPIB_error(C8 *msg, S32 ibsta, S32 iberr, S32 ibcntl)
{
}

void GPIB_update(S64 cur_freq)
{
}

void GPIB_plot(S64 cur_freq)
{
}

void shutdown(void)
{
   ClosePortTalk();
}

#endif

/******************************************/
S32 elapsed_msec(void)
{
   static S64 f = -1;
                                
   if (f == -1)
      {
      QueryPerformanceFrequency((LARGE_INTEGER *) &f);
      }

   S64 i;
   QueryPerformanceCounter  ((LARGE_INTEGER *) &i);

   return S32((i * 1000L) / f);
}                             

/******************************************/
void main(S32 argc, C8 *argv[])
{                	
   U32 i;
   S64 cur_freq,last_freq;
   S64 freq;

   //
   // Enable port access under NT-based versions of Windows
   //

   OpenPortTalk();

   DWORD BytesReturned;        

   DeviceIoControl(PortTalk_Handle,
                   IOCTL_IOPM_ALLOW_EXCUSIVE_ACCESS,
                   NULL,
                   0,    
                   NULL,
                   0,
                  &BytesReturned,
                   NULL);

   DWORD dwProcessId = GetCurrentProcessId();

   DeviceIoControl(PortTalk_Handle,
                   IOCTL_ENABLE_IOPM_ON_PROCESSID,
                  &dwProcessId,
                   4,
                   NULL,
                   0,
                  &BytesReturned,
                   NULL);

   //
   // Display signon banner
   //

   printf("\n");
   printf("Win32 test bed for Stellex YIG PLL synthesizer\n");
   printf("V"VERSION" of "__DATE__" by John Miles, KE5FX (jmiles@pop.net)\n");
   printf("\n\n");

   //
   // Usage: stellex port-address (e.g., 378)
   //

   if (argc == 1)
      {
      printf("Usage: stellex <LPT port address in hexadecimal>\n\nExample: 'stellex 378' for LPT port at address 0x378\n");
      exit(1);
      }

   out_port = strtol(argv[1], NULL, 16);

   printf("PLL using LPT port at address 0x%X\n",out_port);

   atexit(shutdown);

   //
   // Set synthesizer output frequency to default
   //

   cur_freq          = FOUT_DEFAULT;
   last_freq         = -1;
   SA_update         = 1;

   srand(time(NULL));

   //
   // Send data to port
   //

   printf("\n");
   printf("   R: Set random frequency\n");
   printf("   E: Enter frequency in Hz (3- or 4-digit value treated as MHz)\n");
   printf(" a/z: Decrement/increment frequency by %d Hz\n",STEP_SIZE_HZ * PRESCALE_FACTOR);
   printf(" A/Z: Go to frequency lower/upper limit\n");
#if ANALYZER_SUPPORT
   printf(" u/U: Turn spectrum analyzer updates off/on\n");
#if TEK_49X_CONTROL
   printf("   P: Save analyzer plot file for current frequency\n");
#endif
#endif
   printf(" Spc: Force reload of PLL registers\n");
   printf("\n");
   printf(" ESC: Quit\n\n");

   while (1)
      {
      Sleep(10);

      //
      // Create synthesizer object if it does not already exist
      //

      if (synth == NULL)
         {
         synth = new PLL(out_port,
                         PLLCTRL_ADF411X,
                         REF_CLOCK_HZ / STEP_SIZE_HZ,
                         TRUE);
         last_freq = -1;
         }

      //
      // Update frequency
      // 

      if (last_freq != cur_freq)
         {
         //
         // Round frequency to STEP_SIZE_HZ * PRESCALE_FACTOR Hz
         // 

         S64 n = cur_freq / (STEP_SIZE_HZ * PRESCALE_FACTOR);

         cur_freq = n * STEP_SIZE_HZ * PRESCALE_FACTOR;

         last_freq = cur_freq;

         printf("Frequency: %I64d Hz",cur_freq);

         synth->set_frequency(cur_freq / PRESCALE_FACTOR,
                              REF_CLOCK_HZ);

         //                                            
         // See how long it takes the PLL to lock
         //

#if WAIT_FOR_LOCK

         S32 start_acquisition = elapsed_msec();

         S32 poll = 0;

         while (1)
            {
            S32 locked = TRUE;

            for (S32 i=0; i < 100; i++)
               {
               if (synth->locked() != LD_ACTIVE_LEVEL)
                  {
                  locked = FALSE;
                  break;
                  }
               }

            if (locked)
               {
               break;
               }

            poll++;

            if (!(poll % 100000))
               {
               printf(".");

               if (kbhit())
                  {
                  printf("Aborted");
                  getch();
                  break;
                  }
               }
            }

         S32 lock_time = elapsed_msec() - start_acquisition;
         printf(": Lock time = %d ms", lock_time);
#endif

         printf("\n");

         //
         // Tune analyzer to new center frequency
         //

         if (SA_update)
            {
            GPIB_update(cur_freq);
            }
         }

      //
      // Check for keyboard command
      //

      if (kbhit())
         {
         switch (getch())
            {
            //
            // r: Random frequency
            //

            case 'r':
            case 'R':
               {
               cur_freq = rand();
               cur_freq = (cur_freq << 16) + rand();
               cur_freq %= (FOUT_MAX-FOUT_MIN);
               cur_freq += FOUT_MIN;
               break;
               }

            //
            // e: Enter frequency
            //

            case 'e':
            case 'E':
               {
               printf("\nEnter new frequency in Hz: ");

               C8 buff[128];
               gets(buff);

               cur_freq = _atoi64(buff);

               if (cur_freq < 9999)
                  {
                  cur_freq *= 1000000;
                  }

               break;
               }

            //
            // a-z: Adjust frequency by STEP_SIZE_HZ * PRESCALE_FACTOR Hz
            //

            case 'a':
               {
               cur_freq -= (STEP_SIZE_HZ * PRESCALE_FACTOR); 
               break;
               }

            case 'z':
               {
               cur_freq += (STEP_SIZE_HZ * PRESCALE_FACTOR); 
               break;
               }

            //
            // A/Z: Frequency limits
            //

            case 'A':
               cur_freq = FOUT_MIN;
               break;

            case 'Z':
               cur_freq = FOUT_MAX;
               break;

            //
            // Space: Reinitialize synthesizer
            //

            case ' ':

               delete synth;
               synth = NULL;
               break;


            //
            // u/U: SA updates on/off
            // p/P: Take screenshot
            //

#if ANALYZER_SUPPORT
            case 'u':
               SA_update = 0;
               printf("Spectrum analyzer GPIB updates off\n");
               break;

            case 'U':
               SA_update = 1;
               printf("Spectrum analyzer GPIB updates on\n");
               GPIB_update(cur_freq);
               break;

#if TEK_49X_CONTROL
            case 'p':
            case 'P':

               GPIB_plot(cur_freq);
               break;
#endif
#endif

            //
            // ESC: Exit
            //

            case 27:
               exit(0);
            }
         }
      }
}

