/****************************************************************/
/*                                                              */
/* PLLCTRL.CPP: PLL setup routines for Stellex YTO synthesizer  */
/*              under Atmel or Win32-based control              */
/*                                                              */
/* Compilable with AVR-GCC or MSVC                              */
/*                                                              */
/* John Miles, KE5FX (jmiles@pop.net)                           */
/*                                                              */
/****************************************************************/

#ifdef __AVR__
#include <avr\io.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "typedefs.h"      

//
// ADF411x / LMX23x6 PLL port bit assignments
//

#ifdef __AVR__

#define PLL_DATA    (1 << 0)   // Output 
#define PLL_CLK     (1 << 1)   // Output 
#define PLL_LATCH   (1 << 2)   // Output
#define PLL_LOCKED  (1 << 3)   // Input

#else

#define PLL_CLK     0x01
#define PLL_DATA    0x02
#define PLL_LATCH   0x04
#define PLL_LOCKED  0x08

#endif

//
// ADF411x / LMX23x6 PLL latch labels
// 

#define PLL_LATCH_R    0x00
#define PLL_LATCH_N    0x01
#define PLL_LATCH_INIT 0x03

//
// Enums used for PLL chip identification
//

enum PLLCTRL_CHIPTYPE
{
   PLLCTRL_LMX2306,    // National Semiconductor LMX2306
   PLLCTRL_LMX23X6,    // National Semiconductor LMX2316, LMX2326
   PLLCTRL_ADF411X,    // Analog Devices ADF4110, ADF4111, ADF4112, ADF4113
};

//
// PLL class definition
//

class PLL
{
public:

#ifndef __AVR__
   U16              control_port;
#endif
   PLLCTRL_CHIPTYPE chip_type;
   S32              P;
   S32              R;
   S32              CP;
   BOOL             POLARITY;

   // -----------------------------------------------------------
   // Constructor
   // -----------------------------------------------------------
   
   PLL(
#ifndef __AVR__
            S32              _control_port,  // (port is hardwired to PORTB/PINB on AVR)
#endif
            PLLCTRL_CHIPTYPE _chip_type,
            S32              _R,
            BOOL             _POLARITY,
            S32              _P  = 16,
            S32              _CP = 7)
      {
      chip_type = _chip_type;
      R         = _R;
      P         = _P;
      CP        = _CP;
      POLARITY  = _POLARITY;
                
#ifndef __AVR__
      control_port = (U16) _control_port;
#endif
      init();
      }
      
   // -----------------------------------------------------------
   // Initialize PLL
   // -----------------------------------------------------------

   void init(void)
      {       
      //
      // Initialize PLL chip with specified R and P divisors and phase-detector
      // polarity
      //
      // Enable high-precision digital lock detection
      //
      // Prescaler division modulus (P) and charge-pump gain (CP) settings
      // are not specifiable with National Semiconductor parts
      // 

      if (chip_type == PLLCTRL_ADF411X)
         {
         PLL_write_word(24, PLL_LATCH_INIT | (1L << 4)      |   
                                 (POLARITY ? (1L << 7) : 0) |
                                             (CP << 15)     |   
                                    ((log2(P)-3) << 22));
         PLL_latch_data();

         PLL_write_word(24, PLL_LATCH_R | (R << 2) |   
                                         (1L << 20));
         PLL_latch_data();
         }
      else
         {
         PLL_write_word(21, PLL_LATCH_INIT | (1L << 4) |
                                 (POLARITY ? (1L << 7) : 0));
         PLL_latch_data();

         PLL_write_word(21, PLL_LATCH_R | (R << 2) |  
                                         (1L << 20));
         PLL_latch_data();
         }
      }

   // -----------------------------------------------------------
   // Utility functions
   // -----------------------------------------------------------

   static void IO_wait(void)
      {
      //
      // Delay approx. 1 ms
      //

      volatile S16 v = 20;    // 500 for 16-MHz ATmega128

      while (v--)
         {
#ifdef __AVR__
         //
         // Approx. 2 us on 16 MHz ATmega128
         //

         volatile S8 d = 2;

         while (d--);
#else
         //
         // Approx. 2 us on Win98 PC
         //

         _asm
            {
            mov edx,0x80
            out dx,al
            }
#endif
         }
      }

   static S32 log2(S32 val)
      {
      S8 n = -1;

      while (1)
         {
         if (!val)
            {
            break;
            }

         val >>= 1;
         n++;
         }

      return n;
      }

   inline void PORT_write(U8 val)
      {
#ifdef __AVR__
      PORTB = val;
#else
      U32 port = control_port;

      _asm
         {
         mov edx,port
         mov al,val
         out dx,al
         }
#endif
      }

   void BUS_write_word(S8 bits, S32 data, U8 data_line, U8 clk_line)
      {
      //
      // Drop all signals
      //

      PORT_write(0);

      for (S8 i=bits-1; i >= 0; i--)
         {
         //
         // Shift word into input register, starting with MSB
         // After asserting each bit, toggle CLK briefly to load it
         //

         if (data & (1L << i))
            {
            PORT_write(data_line);
            PORT_write(data_line | clk_line);
            PORT_write(data_line);
            }
         else
            {
            PORT_write(0);
            PORT_write(clk_line);
            PORT_write(0);
            }
         }

      IO_wait();
      }

   void PLL_write_word(S32 bits, S32 data)
      {
      BUS_write_word(bits, data, PLL_DATA, PLL_CLK);
      }

   void PLL_latch_data(void)
      {
      //
      // Assert LE briefly to update destination register from input buffer
      //

      PORT_write(PLL_LATCH);
      PORT_write(0);
      }

   // -----------------------------------------------------------
   // Set desired N divisor to determine PLL output frequency
   //
   // Valid output frequencies are those which yield an integral
   // N value for N = hertz / (REF_FREQ / R)
   // -----------------------------------------------------------

   void set_N_divisor(S32 N)
      {
#ifndef __AVR__
      //
      // Calculate modulus values for A and B counters based on N
      //
      // Performance-critical applications may omit the A and B calculations
      // for the National chips
      //

      S32 B = N / P;
      S32 A = N - (B * P);

      //
      // Make sure they're within the limits of the PLL chip
      //

      assert(N >= (P*P)-P);
      assert(B >= A);
      assert(B <= 8191);     
      assert(A >= 0);        

      if (chip_type == PLLCTRL_ADF411X)
         {
         assert(B >= 3);
         assert(A <= 63);       
         }
      else if (chip_type == PLLCTRL_LMX23X6)
         {
         assert(A <= 31);       
         }
      else if (chip_type == PLLCTRL_LMX2306)
         {
         assert(A <= 7);
         }
#endif

      //
      // Program the PLL
      //

      if (chip_type == PLLCTRL_ADF411X)
         {
         PLL_write_word(24, PLL_LATCH_N | (A << 2) | (B << 8));
         PLL_latch_data();
         }
      else
         {
         PLL_write_word(21, PLL_LATCH_N | (N << 2) /* | (1 << 20) */);
         PLL_latch_data();
         }
      }

   // -----------------------------------------------------------
   // Set desired output frequency in Hz
   //
   // Valid output frequencies are those which yield an integral
   // N value for N = hertz / (REF_FREQ / R)
   // -----------------------------------------------------------

   void set_frequency(U64 hertz, //)
                      S32 reference)
      {
      S32 N = S32(hertz / (reference / R));

#ifndef __AVR__
      U64 rounded = U64(N) * (reference / R);
      assert(hertz == rounded);
#endif

      set_N_divisor(N);
      }

   // -----------------------------------------------------------
   // Return TRUE if PLL locked
   // -----------------------------------------------------------

   BOOL locked(void)
      {
#ifdef __AVR__
      return (PINB & PLL_LOCKED) != 0;
#else
      U32 result = 0;
      U32 port = control_port + 1;

      _asm mov edx,port
      _asm xor eax,eax
      _asm in al,dx
      _asm mov result,eax

      return (result & PLL_LOCKED) != 0;
#endif
      }
};
