/****************************************************************/
/*                                                              */
/* STELLAVR.CPP: Stellex synthesizer controller for Atmel chips */
/*                                                              */
/* Compile with AVR-GCC                                         */
/*                                                              */
/* jmiles@pop.net                                               */
/*                                                              */
/****************************************************************/

#define VERSION "1.00"

#include <avr\io.h>
#include <avr\interrupt.h>

#include <stdlib.h>
#include <string.h>

//
// General type definitions for portability
//
// The following types are usable under both MSVC for Win32 and 
// GCC for Atmel 
// 

#ifndef TYPEDEFS_H
#define TYPEDEFS_H

#ifndef US_TYPEDEFS
#define US_TYPEDEFS

#ifndef _WINDOWS_
#ifdef __cplusplus
typedef bool             BOOL;
#else
typedef unsigned char    BOOL;
#endif
#endif
typedef signed   long    BOOL32;
typedef unsigned char    U8 ;
typedef unsigned short   U16;
typedef unsigned long    U32;
typedef          char    C8 ;
typedef signed   char    S8 ;
typedef signed   short   S16;
typedef signed   long    S32;

#ifdef __AVR__
typedef unsigned long long U64;
typedef long long          S64;
#else
#ifdef __WATCOMC__
typedef unsigned __int64 U64;
typedef signed __int64   S64;
#else
typedef unsigned __int64   U64;
typedef   signed __int64   S64;
#endif
#endif

#endif

#ifndef F_TYPEDEFS
#define F_TYPEDEFS

typedef float           SINGLE;
typedef double          DOUBLE;

#endif

#ifndef YES
#define YES 1
#endif

#ifndef NO
#define NO  0
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE  0
#endif

#endif

//
// Synthesizer properties
//

#define REF_CLOCK_HZ    10000000
#define STEP_SIZE_HZ    250000
#define PRESCALE_FACTOR 4      // (Prescaler before LMX2326)

//
// Port definitions
//

#define PLL_DATA   (1 << 0)    // Output 
#define PLL_CLK    (1 << 1)    // Output 
#define PLL_LATCH  (1 << 2)    // Output
#define PLL_LOCKED (1 << 3)    // Input

//
// ADF411x / LMX23x6 PLL latch labels
// 

#define PLL_LATCH_R    0x00
#define PLL_LATCH_N    0x01
#define PLL_LATCH_INIT 0x03

//
// PLL programming constants
//

const S32 R = REF_CLOCK_HZ / STEP_SIZE_HZ; 

// -----------------------------------------------------------
// Support routines
// -----------------------------------------------------------

void IO_wait(void)
{
   //
   // Delay approx. 1 ms
   //

   volatile S16 v = 20;    // (500 for 16-MHz ATmega128, 20 for ATTiny2313 with 8 MHz RC timer 
                           // in CLKDIV8 mode) 
   while (v--)
      {
      volatile S8 d = 2;

      while (d--);
      }
}

void BUS_write_word(S8 bits, S32 data, U8 data_line, U8 clk_line)
{
   //
   // Drop all signals
   //

   PORTB = 0;

   for (S8 i=bits-1; i >= 0; i--)
      {
      //
      // Shift word into input register, starting with MSB
      // After asserting each bit, toggle CLK briefly to load it
      //

      if (data & (1L << i))
         {
         PORTB = data_line;
         PORTB = data_line | clk_line;
         PORTB = data_line;
         }
      else
         {
         PORTB = 0;
         PORTB = clk_line;
         PORTB = 0;
         }
      }

   IO_wait();
}

void PLL_write_word(S32 bits, S32 data)
{
   BUS_write_word(bits, data, PLL_DATA, PLL_CLK);
}

void PLL_latch_data(void)
{
   //
   // Assert LE briefly to update destination register from input buffer
   //

   PORTB = PLL_LATCH;
   PORTB = 0;
}

// -----------------------------------------------------------
// Set desired N divisor to determine PLL output frequency
// -----------------------------------------------------------

void PLL_set_N_divisor(S32 N)
{
   //
   // Set N divisor, low charge-pump gain
   //

   PLL_write_word(21, PLL_LATCH_N | (N << 2) /* | (1 << 20) */);
   PLL_latch_data();
}

// -----------------------------------------------------------
// Since Fcomp is 250 kHz and each step is multiplied by 4
// by the external prescaler, the N divisor is just the frequency in MHz
// -----------------------------------------------------------

void PLL_set_MHz(S32 MHz)
{
   PLL_set_N_divisor(MHz);
}

// -----------------------------------------------------------
// Return TRUE if PLL is locked
// -----------------------------------------------------------

BOOL PLL_locked(void)
{
   return (PINB & PLL_LOCKED) != 0;
}

//******************************************
//
// main()
//
//******************************************

int main(void)
{
   //
   // PB0-PB2 = outputs for LMX2326 control lines
   // PB3     = input for LMX2326 Fo/LD line, pullup enabled
   //

   DDRB  = 0x07;
   PORTB = 0x08;

   //
   // PD0-PD3 = outputs for BCD thumbwheel digit-select lines (LSD=PD0, MSD=PD3)
   // PD4-PD7 = inputs from BCD thumbwheel switches ('1'=PD4, '8'=PD7, with pullups enabled)
   //

   DDRD  = 0x0F;
   PORTD = 0xF0;

   //
   // PA0, PA1 = outputs for lock status LED
   // PA0=red, PA1=green
   //

   DDRA  = 0x03;
   PORTA = 0x03;

   //
   // Wait 1000 ms for hardware and supplies to stabilize
   //

   for (S16 i=0; i < 1000; i++)
      {
      IO_wait();
      }
                      
   //
   // Initialize PLL chip
   //
   //    Enable digital lock detection
   //    Positive VCO tuning characteristic
   //

   PLL_write_word(21, PLL_LATCH_INIT | (1L << 4) | (1L << 7));
   PLL_latch_data();

   //
   //    R divider = 10 MHz / 250 kHz (40)
   //    Lock detection period = five consecutive reference cycles
   //

   PLL_write_word(21, PLL_LATCH_R | (R << 2) | (1L << 20));
   PLL_latch_data();

   S16 last_MHz = -1;
   S16 iteration = 0;

   for (;;)
      {
      //
      // Poll the lock-detect input and flash the LED if it's low
      //

      iteration++;

      if (PLL_locked())
         {
         PORTA = 0x01;                                // solid green
         }
      else
         {
         PORTA = (iteration & 0x200) ? 0x03 : 0x02;   // flashing red
         }

#ifdef HARDWIRED_MHZ

      //
      // If build options included a parameter of the form /DHARDWIRED_MHZ=6000,
      // set that frequency now, and don't try to read any switches
      //

      S16 MHz = HARDWIRED_MHZ;
#else
      //
      // Read BCD thumbwheel switches by grounding each of the four 
      // digits' common terminals in succession, and looking for grounded
      // ports connected to the 1-2-4-8 pins
      //
      // All four switches' 1-2-4-8 pins are connected in parallel through 
      // steering diodes, ensuring that low readings come only from the 
      // digit whose common terminal is active
      //

      S16 MHz = 0;
      
      for (S8 sw=3; sw >= 0; sw--)
         {
         PORTD = ~(1 << sw);
         MHz = (MHz * 10) + (((~PIND) >> 4) & 0x0f);
         }

      //
      // Values less than 2 GHz are added to 10 GHz 
      // for use with X-band YTOs
      //

      if (MHz < 2000) 
         {
         MHz += 10000;
         }
#endif

      //
      // If the frequency didn't change, there's nothing else to do
      //

      if (MHz == last_MHz)
         {
         continue;
         }

      //
      // Otherwise set the new frequency
      //

      PLL_set_MHz(MHz);
      last_MHz = MHz;
      }
}
